# -*- coding: utf-8 -*-
import sys
import operator
import re
import os
import platform
import xmltodict
from collections import defaultdict

"""
Output module for exporting resolved data to one of the supported serialization formats

Author: Amir Zeldes, Siyao Peng
"""

PY3 = sys.version_info[0] == 3


def clean_filename(filename):
	"""
	Heuristically replaces known extensions to create sensible output file name

	:param filename: the input file name to strip extensions from
	"""
	if filename.endswith(".conll10") or filename.endswith(".conllu") and not filename.startswith("."):
		return filename.replace(".conll10", "").replace(".conllu", "")
	else:
		return filename


def output_onto(conll_tokens, markstart_dict, markend_dict, file_name):
	"""
	Outputs analysis results in OntoNotes .coref XML format

	:param conll_tokens: List of all processed ParsedToken objects in the document
	:param markstart_dict: Dictionary from markable starting token ids to Markable objects
	:param markend_dict: Dictionary from markable ending token ids to Markable objects
	:return: serialized XML
	"""

	output_string = '<DOC DOCNO="' + file_name + '">\n<TEXT PARTNO="000">\n'
	for out_tok in conll_tokens:
		if int(out_tok.id) in markstart_dict:
			for out_mark in sorted(markstart_dict[int(out_tok.id)], key=operator.attrgetter('end'), reverse=True):
				output_string += '<COREF ID="' + str(out_mark.group) + '" ENTITY="' + out_mark.entity + '" INFSTAT="' + out_mark.infstat
				if not out_mark.antecedent == "none":
					output_string += '" TYPE="' + out_mark.coref_type
				output_string += '">'
		if int(out_tok.id) > 0:
			output_string += re.sub("&","&amp;",out_tok.text) if ";" not in out_tok.text else out_tok.text
		if int(out_tok.id) in markend_dict:
			for out_mark in markend_dict[int(out_tok.id)]:
				output_string += "</COREF>"
		if int(out_tok.id) > 0:
			output_string += ' '


	return output_string + "\n</TEXT>\n</DOC>\n"


def output_SGML(conll_tokens, markstart_dict, markend_dict):
	"""
	Outputs analysis results as CWB SGML (with nesting), one token per line and markables in <referent> tags

	:param conll_tokens: List of all processed ParsedToken objects in the document
	:param markstart_dict: Dictionary from markable starting token ids to Markable objects
	:param markend_dict: Dictionary from markable ending token ids to Markable objects
	:return: serialized SGML
	"""

	output_string = ""
	for out_tok in conll_tokens:
		if int(out_tok.id) in markstart_dict:
			for out_mark in sorted(markstart_dict[int(out_tok.id)], key=operator.attrgetter('end'), reverse=True):
				output_string += '<referent id="' + str(out_mark.id) + '" entity="' + out_mark.entity + '" group="' + str(out_mark.group)
				if not out_mark.antecedent == "none":
					output_string += '" antecedent="' + str(out_mark.antecedent.id) + '" type="' + out_mark.coref_type
				output_string += '">\n'
		if int(out_tok.id) > 0:
			output_string += out_tok.text + "\n"
		if int(out_tok.id) in markend_dict:
			for out_mark in markend_dict[int(out_tok.id)]:
				output_string += "</referent>\n"

	return output_string


def output_conll(conll_tokens, markstart_dict, markend_dict, file_name, output_infstat=False):
	"""
	Outputs analysis results in CoNLL format, one token per line and markables with opening
	and closing numbered brackets. Compatible with CoNLL scorer.

	:param conll_tokens: List of all processed ParsedToken objects in the document
	:param markstart_dict: Dictionary from markable starting token ids to Markable objects
	:param markend_dict: Dictionary from markable ending token ids to Markable objects
	:param file_name: name of the source file (dependency data) to create header for CoNLL file
	:param output_infstat: whether to append the infstat property of each markable in a separate column (default False)
	:return: serialized conll format in plain text
	"""
	output_string = "# begin document " + str(file_name).replace(".conll10", "").replace("_xrenner","").replace("_hyph","").replace("_deped","").replace("_decyc","")+"\n"
	i = -1
	for out_tok in conll_tokens[1:]:
		i += 1
		coref_col = ""
		line = str(i) + "\t" + out_tok.text + "\t"
		infstat_col = ""
		if output_infstat:
			infstat_col = "_\t"
		if int(out_tok.id) in markstart_dict:
			for out_mark in sorted(markstart_dict[int(out_tok.id)], key=operator.attrgetter('end'), reverse=True):
				coref_col += "(" + str(out_mark.group)
				if output_infstat:
					infstat_col = out_mark.infstat + "\t"
				if int(out_tok.id) in markend_dict:
					if out_mark in markend_dict[int(out_tok.id)]:
						coref_col += ")"
						markend_dict[int(out_tok.id)].remove(out_mark)
		if int(out_tok.id) in markend_dict:
			for out_mark in markend_dict[int(out_tok.id)]:
				if out_mark in markstart_dict[int(out_tok.id)]:
					coref_col += ")"
				else:
					if len(coref_col) > 0:
						if coref_col[-1].isdigit():
							coref_col += "|"  # Use pipe to separate group 1 opening and 2 closing leading to (12) -> (1|2)
					coref_col += str(out_mark.group) + ")"
		if int(out_tok.id) not in markstart_dict and int(out_tok.id) not in markend_dict:
			coref_col = "_"

		line += infstat_col + coref_col
		output_string += line + "\n"
	output_string += "# end document\n\n"
	return output_string


def output_conll_sent(conll_tokens, markstart_dict, markend_dict, file_name, output_infstat=False, output_entity=True):
	"""
	Outputs analysis results in CoNLL format, one token per line and markables with opening
	and closing numbered brackets. Compatible with CoNLL scorer.

	:param conll_tokens: List of all processed ParsedToken objects in the document
	:param markstart_dict: Dictionary from markable starting token ids to Markable objects
	:param markend_dict: Dictionary from markable ending token ids to Markable objects
	:param file_name: name of the source file (dependency data) to create header for CoNLL file
	:param output_infstat: whether to append the infstat property of each markable in a separate column (default False)
	:param output_entity: whether to output entity types in coref ID column (default True)
	:return: serialized conll format in plain text
	"""
	output_string = "# begin document " + str(file_name).replace(".conll10", "") + "\n"
	i = -1
	current_sent = ""
	for out_tok in conll_tokens[1:]:
		if current_sent != out_tok.sentence.sent_num:
			current_sent = out_tok.sentence.sent_num
			output_string += "\n"
			i = 0

		i += 1
		coref_col = ""
		line = str(i) + "\t" + out_tok.text + "\t"
		infstat_col = ""
		if int(out_tok.id) in markstart_dict:
			for out_mark in sorted(markstart_dict[int(out_tok.id)], key=operator.attrgetter('end'), reverse=True):
				coref_col += "(" + str(out_mark.group)
				if output_entity:
					coref_col += "-" + out_mark.entity
				if output_infstat:
					infstat_col = out_mark.infstat
				if int(out_tok.id) in markend_dict:
					if out_mark in markend_dict[int(out_tok.id)]:
						coref_col += ")"
						markend_dict[int(out_tok.id)].remove(out_mark)
		if int(out_tok.id) in markend_dict:
			for out_mark in markend_dict[int(out_tok.id)]:
				if out_mark in markstart_dict[int(out_tok.id)]:
					coref_col += ")"
				else:
					if len(coref_col) > 0:
						if coref_col[-1].isdigit():
							coref_col += "|"  # Use pipe to separate group 1 opening and 2 closing leading to (12) -> (1|2)
					coref_col += str(out_mark.group)
					if output_entity:
						coref_col += "-" + out_mark.entity
					coref_col += ")"
		if int(out_tok.id) not in markstart_dict and int(out_tok.id) not in markend_dict:
			coref_col = "_"

		line += infstat_col + "\t" + coref_col
		output_string += line + "\n"
	output_string += "# end document\n\n"
	return output_string


def output_HTML(conll_tokens, markstart_dict, markend_dict, rtl=False):
	"""
	Outputs analysis results as HTML (assuming jquery, xrenner css and js files), one token per line and
	markables in <div> tags with Font Awesome icons and colored groups.

	:param conll_tokens: List of all processed ParsedToken objects in the document
	:param markstart_dict: Dictionary from markable starting token ids to Markable objects
	:param markend_dict: Dictionary from markable ending token ids to Markable objects
	:return: serialized HTML
	"""

	if rtl:
		rtl_style = ' style="direction: rtl"'
	else:
		rtl_style = ""

	output_string = '''<html>
<head>
	<link rel="stylesheet" href="http://corpling.uis.georgetown.edu/xrenner/css/renner.css" type="text/css" charset="utf-8"/>
	<link rel="stylesheet" href="https://corpling.uis.georgetown.edu/xrenner/css/font-awesome-4.2.0/css/font-awesome.min.css"/>
	<meta http-equiv="content-type" content="text/html; charset=utf-8"/>
</head>
<body'''+rtl_style+'''>
<script src="http://corpling.uis.georgetown.edu/xrenner/script/jquery-1.11.3.min.js"></script>
<script src="http://corpling.uis.georgetown.edu/xrenner/script/chroma.min.js"></script>
<script src="http://corpling.uis.georgetown.edu/xrenner/script/xrenner.js"></script>
'''
	for out_tok in conll_tokens:
		if int(out_tok.id) in markstart_dict:
			for out_mark in sorted(markstart_dict[int(out_tok.id)], key=operator.attrgetter('end'), reverse=True):
				info_string = "class: " + str(out_mark.entity) + " | subclass: " + str(out_mark.subclass) + \
				              "&#10;definiteness: " + str(out_mark.definiteness) + " | agree: " + str(out_mark.agree) + \
				              "&#10;cardinality: " + str(out_mark.cardinality) + " | form: "+ str(out_mark.form) + \
				              "&#10;func: " + str(out_mark.func) + \
				              "&#10;core_text: " + str(out_mark.core_text) + " | lemma: " + str(out_mark.lemma)
				if out_mark.speaker != "":
					info_string += "&#10;speaker: " + out_mark.speaker
				if not out_mark.antecedent == "none":
					info_string += '&#10;coref_type: ' + out_mark.coref_type
				if "matching_rule" in out_mark.__dict__:
					info_string += "&#10;coref_rule: " + out_mark.matching_rule
				output_string += '<div id="' + out_mark.id + '" head="' + out_mark.head.id + '" onmouseover="highlight_group(' + \
				"'" + str(out_mark.group) + "'" + ')" onmouseout="unhighlight_group(' + "'" + str(out_mark.group) + "'" + ')" class="referent" group="' + str(out_mark.group) + '" title="' + info_string
				if not out_mark.antecedent == "none":
					output_string += '" antecedent="' + out_mark.antecedent.id
				output_string += '"><span class="entity_type">' + get_glyph(out_mark.entity) + '</span>\n'
		if int(out_tok.id) > 0:
			output_string += out_tok.text.replace("-RRB-", ")").replace("-LRB-", "(").replace("-LSB-", "[").replace("-RSB-", "]") + "\n"
		if int(out_tok.id) in markend_dict:
			for out_mark in markend_dict[int(out_tok.id)]:
				output_string += "</div>\n"
	output_string += '<script>colorize();</script>\n'
	output_string += '''</body>
</html>'''
	return output_string


def output_PAULA(conll_tokens, markstart_dict, markend_dict, docname, docpath):
	"""
	Outputs analysis results as PAULA standoff XML. Separate files for tokens, markables and coreference links
	plus annotations. This format is the most complete, distinguishing apposition, anaphora, cataphora and other
	coreference types as edge annotations.

	:param conll_tokens: List of all processed ParsedToken objects in the document
	:param markstart_dict: Dictionary from markable starting token ids to Markable objects
	:param markend_dict: Dictionary from markable ending token ids to Markable objects
	:return: void
	"""

	paula_text = ""
	paula_tokens = ""
	paula_markables = ""
	paula_entities = ""
	paula_rels = ""
	paula_rel_annos = ""
	cursor = 1
	rel_id = 1
	paula_text_header = '''<?xml version="1.0" standalone="no"?>
<!DOCTYPE paula SYSTEM "paula_text.dtd">

<paula version="1.0">
<header paula_id="renner.out_text" type="text"/>

<body>
'''
	paula_tok_header = '''<?xml version="1.0" standalone="no"?>

<!DOCTYPE paula SYSTEM "paula_mark.dtd">
<paula version="1.0">

<header paula_id="renner.out_tok"/>

<markList xmlns:xlink="http://www.w3.org/1999/xlink" type="tok" xml:base="xrenner.''' + docname + '''.text.xml">
'''

	paula_mark_header = '''<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<!DOCTYPE paula SYSTEM "paula_mark.dtd">
<paula version="1.0">

<header paula_id="xrenner.''' + docname + '''_referentSeg"/>

<markList xmlns:xlink="http://www.w3.org/1999/xlink" type="referentSeg" xml:base="xrenner.''' + docname + '''.tok.xml">
'''

	paula_entity_header = '''<?xml version="1.0" encoding="UTF-8" standalone="no"?>

<!DOCTYPE paula SYSTEM "paula_feat.dtd">
<paula version="1.0">

<header paula_id="xrenner.''' + docname + '''_referentSeg_entity"/>

<featList xmlns:xlink="http://www.w3.org/1999/xlink" type="entity" xml:base="xrenner.''' + docname + '''.referentSeg.xml">
'''

	paula_coref_header = '''<?xml version="1.0" standalone="no"?>

<!DOCTYPE paula SYSTEM "paula_rel.dtd">
<paula version="1.0">

<header paula_id="xrenner.''' + docname + '''.referentSeg_coref"/>

<relList xmlns:xlink="http://www.w3.org/1999/xlink" type="coref" xml:base="xrenner.''' + docname + '''.referentSeg.xml">
'''
	paula_coref_type_header = '''<?xml version="1.0" encoding="UTF-8" standalone="no"?>

<!DOCTYPE paula SYSTEM "paula_feat.dtd">
<paula version="1.0">

<header paula_id="xrenner.''' + docname + '''.referentSeg_coref_type"/>

<featList xmlns:xlink="http://www.w3.org/1999/xlink" type="type" xml:base="xrenner.''' + docname + '''.referentSeg_coref.xml">
'''

	if not os.path.exists(docpath + os.sep + docname):
		os.makedirs(docpath + os.sep + docname)
	elif os.path.isfile(docpath + os.sep + docname):
		raise("Unable to create document directory. There is already a file named " + docpath + os.sep + docname + "\n")

	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.text.xml', 'w')
	f.write(paula_text_header)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.tok.xml', 'w')
	f.write(paula_tok_header)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg.xml', 'w')
	f.write(paula_mark_header)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg_entity.xml', 'w')
	f.write(paula_entity_header)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg_coref.xml', 'w')
	f.write(paula_coref_header)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg_coref_type.xml', 'w')
	f.write(paula_coref_type_header)

	del conll_tokens[0]
	for out_tok in conll_tokens:
		paula_text += out_tok.text + " "
		if int(out_tok.id) in markstart_dict:
			for out_mark in markstart_dict[int(out_tok.id)]:
				if out_mark.end > out_mark.start:
					paula_markables += '<mark id="' + out_mark.id + '"  xlink:href="#xpointer(id(' + "'tok_" + str(out_mark.start) + "')/range-to(id('tok_" + str(out_mark.end) + "')))" + '"><!-- ' + out_mark.text + " -->\n"
				else:
					paula_markables += '<mark id="' + out_mark.id + '"  xlink:href="#tok_' + str(out_mark.start) + '"><!-- ' + out_mark.text + " -->\n"
				paula_entities += '<feat xlink:href="#' + out_mark.id + '" value="' + out_mark.entity + '"><!-- ' + out_mark.text + " -->\n"
				if out_mark.antecedent != "none":
					paula_rels += '<rel id="rel_' + str(rel_id) + '" xlink:href="#' + out_mark.id + '" target="#' + out_mark.antecedent.id + '"/><!-- ' + out_mark.text + ' ... ' + out_mark.antecedent.text + ' -->\n'
					paula_rel_annos += '<feat xlink:href="#rel_' + str(rel_id) + '" value="' + out_mark.coref_type + '"/><!-- ' + out_mark.text + ' ... ' + out_mark.antecedent.text + ' -->\n'
					rel_id += 1

		paula_tokens += '<mark id="tok_' + out_tok.id + '" xlink:href="#xpointer(string-range(//body,' + "'', " + str(cursor) + "," + str(len(out_tok.text)) + "))" + '"/><!-- ' + out_tok.text + " -->\n"
		cursor += len(out_tok.text) + 1

	paula_text += "\n</body>\n</paula>\n"
	paula_tokens += "</markList>\n</paula>\n"
	paula_markables += "</markList>\n</paula>\n"
	paula_entities += "</featList>\n</paula>\n"
	paula_rels += "</relList>\n</paula>\n"
	paula_rel_annos += "</featList>\n</paula>\n"
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.text.xml', 'a')
	f.write(paula_text)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.tok.xml', 'a')
	f.write(paula_tokens)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg.xml', 'a')
	f.write(paula_markables)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg_entity.xml', 'a')
	f.write(paula_entities)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg_coref.xml', 'a')
	f.write(paula_rels)
	f = open(docpath + os.sep + docname + os.sep + 'xrenner.' + docname + '.referentSeg_coref_type.xml', 'a')
	f.write(paula_rel_annos)


def output_webanno(conll_tokens, markables):
	output = '''<?xml version="1.0" encoding="UTF-8"?>
	<xmi:XMI xmlns:cas="http:///uima/cas.ecore"
	    xmlns:type2="http:///de/tudarmstadt/ukp/dkpro/core/api/metadata/type.ecore"
	    xmlns:dependency="http:///de/tudarmstadt/ukp/dkpro/core/api/syntax/type/dependency.ecore"
	    xmlns:type5="http:///de/tudarmstadt/ukp/dkpro/core/api/syntax/type.ecore"
	    xmlns:type3="http:///de/tudarmstadt/ukp/dkpro/core/api/ner/type.ecore"
	    xmlns:custom="http:///webanno/custom.ecore"
	    xmlns:type4="http:///de/tudarmstadt/ukp/dkpro/core/api/segmentation/type.ecore"
	    xmlns:tcas="http:///uima/tcas.ecore"
	    xmlns:tweet="http:///de/tudarmstadt/ukp/dkpro/core/api/lexmorph/type/pos/tweet.ecore"
	    xmlns:chunk="http:///de/tudarmstadt/ukp/dkpro/core/api/syntax/type/chunk.ecore"
	    xmlns:xmi="http://www.omg.org/XMI"
	    xmlns:type="http:///de/tudarmstadt/ukp/dkpro/core/api/coref/type.ecore"
	    xmlns:morph="http:///de/tudarmstadt/ukp/dkpro/core/api/lexmorph/type/morph.ecore"
	    xmlns:constituent="http:///de/tudarmstadt/ukp/dkpro/core/api/syntax/type/constituent.ecore"
	    xmlns:pos="http:///de/tudarmstadt/ukp/dkpro/core/api/lexmorph/type/pos.ecore" xmi:version="2.0">
	    <cas:NULL xmi:id="0"/>
	    <cas:Sofa xmi:id="12000" sofaNum="1" sofaID="_InitialView" mimeType="text"
	        sofaString="'''

	text_string = ""
	all_ids_string = ""
	text_length = 0
	for token in conll_tokens:
		escape_offset = 0
		escaped_token = token.text
		if '&' in escaped_token:
			escaped_token = escaped_token.replace('&', "&amp;")
			escape_offset += 4 * token.text.count("&")
		if '"' in escaped_token:
			escaped_token = escaped_token.replace('"',"&quot;")
			escape_offset += 5 * token.text.count('"')
		if '>' in escaped_token:
			escaped_token = escaped_token.replace('>', "&gt;")
			escape_offset += 3 * token.text.count(">")
		if '<' in escaped_token:
			escaped_token = escaped_token.replace('<', "&lt;")
			escape_offset += 3 * token.text.count("<")

		text_string += escaped_token + " "
		if PY3:
			text_length += len(token.text) + 1
		else:
			text_length += len(token.text.decode("utf-8")) + 1

	output += text_string
	output += '"/>\n<type2:DocumentMetaData xmi:id="10001" sofa="12000" begin="0" end="' + str(text_length - 1) + '"'
	output += '''
		language="x-unspecified"
        documentTitle="renner_out.tcf" documentId="renner_out"
        documentUri="file:/srv/webanno/repository/project/2/document/4/source/renner_out.tcf"
        collectionId="file:/srv/webanno/repository/project/2/document/4/source/"
        documentBaseUri="file:/srv/webanno/repository/project/2/document/4/source/"
        isLastSegment="false"/>\n'''

	cursor = 0
	current_sent = 1
	sent_begin = 0
	sentences_string = ""
	tok_starts = []
	tok_ends = []

	for token in conll_tokens:
		if PY3:
			token_text = token.text
		else:
			token_text = token.text.decode("utf-8")


		output += '\t<type4:Token xmi:id="' + str(int(token.id) + 1) + '" sofa="12000" begin="' + str(cursor) + '" end="' + str(cursor + len(token_text)) + '"/>\n'
		all_ids_string += str(int(token.id) + 1) + " "
		tok_starts.append(cursor)
		tok_ends.append(cursor + len(token_text))

		if token.sentence.sent_num > current_sent:
			sentences_string += '\t<type4:Sentence xmi:id="' + str(4000 + current_sent) + '" sofa="12000" begin="' + str(sent_begin) + '" end="' + str(cursor - 1) + '"/>\n'
			all_ids_string += str(4000 + current_sent) + " "
			current_sent += 1
			sent_begin = cursor
		cursor += len(token_text) + 1

	sentences_string += '\t<type4:Sentence xmi:id="' + str(4000 + current_sent) + '" sofa="12000" begin="' + str(sent_begin) + '" end="' + str(cursor - 1) + '"/>\n'
	all_ids_string += str(4000 + current_sent) + " "

	output += sentences_string

	mark_counter = 1
	mark_xmi_ids = {}
	for mark in markables:
		output += '\t<custom:Referent xmi:id="' + str(5000 + mark_counter) + '" sofa="12000" begin="' + str(tok_starts[mark.start - 1]) + \
		          '" end="' + str(tok_ends[mark.end - 1]) + '" entity="' + mark.entity + '" infstat="' + mark.infstat + '"/>\n'
		all_ids_string += str(5000 + mark_counter) + " "
		mark_xmi_ids[mark.id] = str(5000 + mark_counter)
		mark_counter += 1

	link_counter = 1
	for mark in markables:
		if mark.antecedent != "none":
			output += '\t<custom:Coref xmi:id="' + str(6000 + link_counter) + '" sofa="12000" begin="' + \
			             str(min(tok_starts[mark.start - 1], tok_starts[int(mark.antecedent.start)-1])) + '" end="' + \
			             str(max(tok_ends[mark.end-1], tok_starts[int(mark.antecedent.end)-1])) + '" Dependent="' + mark_xmi_ids[mark.antecedent.id] \
			             + '" Governor="' + mark_xmi_ids[mark.id] + '" type="' + mark.coref_type + '"/>\n'
			all_ids_string += str(6000 + link_counter) + " "
			link_counter += 1

	output += '''    <type2:TagsetDescription xmi:id="15571" sofa="12000" begin="0" end="0"
        layer="de.tudarmstadt.ukp.dkpro.core.api.syntax.type.dependency.Dependency" name="Tiger"/>
    <type2:TagsetDescription xmi:id="15578" sofa="12000" begin="0" end="0"
        layer="de.tudarmstadt.ukp.dkpro.core.api.ner.type.NamedEntity" name="NER_WebAnno"/>
    <type2:TagsetDescription xmi:id="15585" sofa="12000" begin="0" end="0"
        layer="de.tudarmstadt.ukp.dkpro.core.api.lexmorph.type.pos.POS" name="STTS"/>
		<type2:TagsetDescription xmi:id="15592" sofa="12000" begin="0" end="0" layer="webanno.custom.Coref"
			name="coref_tags"/>
		<type2:TagsetDescription xmi:id="15599" sofa="12000" begin="0" end="0"
			layer="webanno.custom.Referent" name="infstat_tags"/>\n'''

	output += '<cas:View sofa="12000" members="' + all_ids_string.strip() + '"/>\n</xmi:XMI>\n'

	return output


def output_webannotsv(conll_tokens, markables, output_infstat=True):
	webannoxmi = xmltodict.parse(output_webanno(conll_tokens, markables))

	if not output_infstat:
		output = ['#FORMAT=WebAnno TSV 3.2',
			'#T_SP=webanno.custom.Referent|entity',
			'#T_RL=webanno.custom.Coref|type|BT_webanno.custom.Referent',
		'', '']
	else:
		output = ['#FORMAT=WebAnno TSV 3.2',
			'#T_SP=webanno.custom.Referent|entity|infstat',
			'#T_RL=webanno.custom.Coref|type|BT_webanno.custom.Referent',
		'', '']

	tokenstring = webannoxmi['xmi:XMI']['cas:Sofa']['@sofaString']

	# ref_id: (ref_start_char, ref_end_char, ref_start_tok, ref_end_tok)
	refdict = defaultdict(lambda: [None, None, None, None])

	singletokens = []

	for sent in webannoxmi['xmi:XMI']['type4:Sentence']:
		sent_id = int(sent['@xmi:id']) - 4000
		sent_start_char = int(sent['@begin'])
		sent_end_char = int(sent['@end'])

		tok_id = 1
		for tok in webannoxmi['xmi:XMI']['type4:Token']:
			tok_start_char = int(tok['@begin'])
			tok_end_char = int(tok['@end'])

			if tok_start_char > sent_end_char:
				break
			elif tok_start_char >= sent_start_char:
				line_ref_string = ''
				line_type_string = ''
				line_coref_string = ''
				line_coref_chain = ''

				if tok_id==1:
					output.append('#Text=%s' % (tokenstring[sent_start_char:sent_end_char]))

				line_output = ['%d-%d' % (sent_id, tok_id), '%d-%d' % (tok_start_char, tok_end_char), tokenstring[tok_start_char:tok_end_char]]


				for ref in webannoxmi['xmi:XMI']['custom:Referent']:

					refdict[int(ref['@xmi:id'])-5000][0] =  ref['@begin']
					refdict[int(ref['@xmi:id']) - 5000][1] = ref['@end']

					if tok_start_char >= int(ref['@begin']) and tok_end_char<= int(ref['@end']):

						if tok_start_char == int(ref['@begin']) and tok_end_char == int(ref['@end']):
							line_ref_string += '%s|' % (ref['@entity'])
							if output_infstat:
								line_type_string += '%s|' % (ref['@infstat'])
						else:
							line_ref_string += '%s[%d]|' % (ref['@entity'], int(ref['@xmi:id'])-5000)
							if output_infstat:
								line_type_string += '%s[%d]|' % (ref['@infstat'], int(ref['@xmi:id'])-5000)

						if tok_start_char == int(ref['@begin']):
							refdict[int(ref['@xmi:id']) - 5000][2] = '%d-%d' % (sent_id, tok_id)

						if tok_end_char == int(ref['@end']):
							refdict[int(ref['@xmi:id']) - 5000][3] = '%d-%d' % (sent_id, tok_id)

						if 'custom:Coref' in webannoxmi["xmi:XMI"]:
							if not isinstance(webannoxmi['xmi:XMI']['custom:Coref'],list):
								webannoxmi['xmi:XMI']['custom:Coref'] = [webannoxmi['xmi:XMI']['custom:Coref']]
							for coref in webannoxmi['xmi:XMI']['custom:Coref']:
								if int(coref['@begin']) == int(ref['@begin']):

									if tok_start_char == int(ref['@begin']):
										# line_coref_chain += '%d[%d_%d]|' % (int(ref['@xmi:id'])-5000, int(coref['@Governor'])-5000, int(coref['@Dependent'])-5000)
										line_coref_chain += '%d[%d_%d]|' % (
										int(coref['@Governor']) - 5000, int(coref['@Governor']) - 5000, int(coref['@Dependent']) - 5000)
										line_coref_string += '%s|' % (coref['@type'])


										if tok_start_char == int(ref['@begin']) and tok_end_char == int(ref['@end']):
											singletokens.append(int(ref['@xmi:id']) - 5000)

				if line_ref_string == '':
					line_ref_string = '_'
				elif line_ref_string.endswith('|'):
					line_ref_string = line_ref_string[:-1]

				if line_type_string == '':
					line_type_string = '_'
				elif line_type_string.endswith('|'):
					line_type_string = line_type_string[:-1]

				if line_coref_chain == '':
					line_coref_chain = '_'
				elif line_coref_chain.endswith('|'):
					line_coref_chain = line_coref_chain[:-1]


				if line_coref_string == '':
					line_coref_string = '_'
				elif line_coref_string.endswith('|'):
					line_coref_string = line_coref_string[:-1]


				line_output += [line_ref_string, line_type_string, line_coref_string, line_coref_chain]

				output.append(line_output)
				tok_id += 1

		# output += '\n'
		output.append('')

	# put the coref token back in
	# TODO: which exact token should the Webanno corefer to ? start, end, or head?

	for id_l, l in enumerate(output):
		if isinstance(l, list):
			corefcol = l[-1]

			if corefcol != '_':
				chains = [re.split(r'[\[\]_]', x) for x in corefcol.split('|')]

				for id_j in range(len(chains)):
					tokenplace = refdict[int(chains[id_j][0])][2]

					if refdict[int(chains[id_j][1])][2] == refdict[int(chains[id_j][1])][3]:
						leftid = '0'
					else:
						leftid = chains[id_j][1]

					if refdict[int(chains[id_j][2])][2] == refdict[int(chains[id_j][2])][3]:
						rightid = '0'
					else:
						rightid = chains[id_j][2]

					if rightid == '0' and leftid == '0':
						chains[id_j] = tokenplace
					else:
						chains[id_j] = tokenplace + '[' + leftid + '_' + rightid + ']'

				output[id_l][-1] = '|'.join(chains)

			output[id_l] = '\t'.join(output[id_l])

	output = '\n'.join(output)

	return output


def get_glyph(entity_type):
	"""
	Generates appropriate Font Awesome icon strings based on entity type strings, such as
	a person icon (fa-male) for the 'person' entity, etc.

	:param entity_type: String specifying the entity type to be visualized
	:return: HTML string with the corresponding Font Awesome icon
	"""
	if entity_type == "person":
		return '<i title="' + entity_type + '" class="fa fa-male"></i>'
	elif entity_type == "place":
		return '<i title="' + entity_type + '" class="fa fa-map-marker"></i>'
	elif entity_type == "time":
		return '<i title="' + entity_type + '" class="fa fa-clock-o"></i>'
	elif entity_type == "abstract":
		return '<i title="' + entity_type + '" class="fa fa-cloud"></i>'
	elif entity_type == "quantity":
		return '<i title="' + entity_type + '" class="fa fa-sort-numeric-asc"></i>'
	elif entity_type == "organization":
		return '<i title="' + entity_type + '" class="fa fa-bank"></i>'
	elif entity_type == "object":
		return '<i title="' + entity_type + '" class="fa fa-cube"></i>'
	elif entity_type == "event":
		return '<i title="' + entity_type + '" class="fa fa-bell-o"></i>'
	elif entity_type == "animal":
		return '<i title="' + entity_type + '" class="fa fa-paw"></i>'
	elif entity_type == "plant":
		return '<i title="' + entity_type + '" class="fa fa-pagelines"></i>'
	elif entity_type == "substance":
		return '<i title="' + entity_type + '" class="fa fa-flask"></i>'
	else:
		return '<i title="' + entity_type + '" class="fa fa-question"></i>'




import re
from .xrenner_marker import remove_suffix_tokens
from .xrenner_propagate import *
from .xrenner_classes import Markable, ParsedToken
from collections import OrderedDict

"""
Module for checking compatibility of various features between markables

Author: Amir Zeldes
"""


def entities_compatible(mark1, mark2, lex):
	"""
	Checks if the entity property of two markables is compatible for possible coreference

	:param mark1: the first of two markables to compare entities
	:param mark2: the second of two markables to compare entities
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	if mark1.entity == mark2.entity:
		return True
	elif mark1.entity is None or mark2.entity is None or mark1.entity == "" or mark2.entity == "":
		return True
	if mark1.form == "pronoun" and (not (mark1.entity == lex.filters["person_def_entity"] and mark2.entity != lex.filters["person_def_entity"]) or mark1.entity_certainty == ''):
		return True
	if mark1.entity != mark2.entity:
		if mark1.entity in mark2.alt_entities and (mark2.entity_certainty == "uncertain" or mark2.entity_certainty == "propagated"):
			return True
		elif mark2.entity in mark1.alt_entities and (mark1.entity_certainty == "uncertain" or mark1.entity_certainty == "propagated"):
			return True
		elif mark2.entity == lex.filters["default_entity"] and mark2.entity_certainty in ["","propagated","uncertain"]:
			return True
		elif mark1.entity == lex.filters["default_entity"] and mark1.entity_certainty in ["","propagated","uncertain"]:
			return True

	return False


def cardinality_compatible(mark1,mark2,lex):
	if "ablations" in lex.debug:
		if "no_cardinality" in lex.debug["ablations"]:
			return True
	if mark1.cardinality!=0:
		if mark2.cardinality!=0:
			if mark1.cardinality != mark2.cardinality:
				return False
	return True


def modifiers_compatible(markable, candidate, lex, allow_force_proper_mod_match=True):
	"""
	Checks whether the dependents of two markables are compatible for possible coreference

	:param markable: :class:`Markable` one of two markables to compare dependents for
	:param candidate: :class:`Markable` the second markable, which is a candidate antecedent for the other markable
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	if markable.id+"|"+candidate.id in lex.incompatible_mod_pairs:
		return False

	if allow_force_proper_mod_match:
		proper_mod_must_match = lex.filters["proper_mod_must_match"]
	else:
		proper_mod_must_match = False

	if not cardinality_compatible(markable,candidate,lex):
		return False

	# Do strict 'no new modifiers' check if desired
	if lex.filters["no_new_modifiers"]:
		first_mark = candidate
		second_mark = markable
		if markable.start < candidate.start:
			first_mark = markable
			second_mark = candidate
		first_mods = (comp_mod.text for comp_mod in first_mark.head.modifiers)
		for mod in second_mark.head.modifiers:
			if lex.filters["det_func"].match(mod.func) is None:  # Exclude determiners from this check
				if mod.text not in first_mods:
					if lex.filters["use_new_modifier_exceptions"]:
						if mod.text not in lex.exceptional_new_modifiers:
							return False
					else:
						return False

	# Check if markable and candidate have modifiers that are in the antonym list together,
	# e.g. 'the good news' should not be coreferent with 'the bad news',
	antonym_check = True
	if "ablations" in lex.debug:
		if "no_antonyms" in lex.debug["ablations"]:
			antonym_check = False
	if antonym_check:
		for mod in markable.head.modifiers:
			if mod.text.lower() in lex.antonyms:
				for candidate_mod in candidate.head.modifiers:
					if candidate_mod.text.lower() in lex.antonyms[mod.text.lower()]:
						markable.non_antecdent_groups.add(candidate.group)
						return False
			elif mod.lemma.lower() in lex.antonyms:
				for candidate_mod in candidate.head.modifiers:
					if candidate_mod.lemma.lower() in lex.antonyms[mod.lemma.lower()]:
						markable.non_antecdent_groups.add(candidate.group)
						return False
			# Check that the two markables do not have non-identical proper noun modifiers
			if proper_mod_must_match:
				if lex.filters["proper_pos"].match(mod.pos):
					candidate_proper_mod_texts = []
					for mod2 in candidate.head.modifiers:
						if lex.filters["proper_pos"].match(mod2.pos):
							candidate_proper_mod_texts.append(mod2.text)
					if mod.text not in candidate_proper_mod_texts and len(candidate_proper_mod_texts) > 0:
						return False

	# Check if markable and candidate have modifiers that are different place names
	# e.g. 'Georgetown University' is incompatible with 'Boston University' even if those entities are not in lexicon
	for mod in markable.head.modifiers:
		if mod.text in lex.entities and (mod.text.istitle() or not lex.filters["cap_names"]):
			if re.sub('\t.*', "", lex.entities[mod.text][0]) == lex.filters["place_def_entity"]:
				for candidate_mod in candidate.head.modifiers:
					if candidate_mod.text != mod.text:
						if candidate_mod.text in lex.entities and (candidate_mod.text.istitle() or not lex.filters["cap_names"]):
							if re.sub('\t.*', "", lex.entities[candidate_mod.text][0]) == lex.filters["place_def_entity"]:
								markable.non_antecdent_groups.add(candidate.group)
								return False

	# Check for each possible pair of modifiers with identical function in the ident_mod list whether they are identical,
	# e.g. for the num function 'the four children' shouldn't be coreferent with 'five other children'
	for mod in markable.head.modifiers:
		for candidate_mod in candidate.head.modifiers:
			# TODO: add support for ident_mod pos func combo:
			# if lex.filters["ident_mod_func"].match(mod.func+"+"+mod.pos) and lex.filters["ident_mod_func"].match(candidate_mod.func+"+"+candidate_mod.pos) and
			# mod.text.lower != candidate_mod.text.lower():
			if lex.filters["ident_mod_func"].match(mod.func) is not None and lex.filters["ident_mod_func"].match(candidate_mod.func) is not None and mod.text.lower != candidate_mod.text.lower():
				markable.non_antecdent_groups.add(candidate.group)
				return False

	# Check that heads are not antonyms themselves
	if markable.head.lemma in lex.antonyms:
		if candidate.head.lemma in lex.antonyms[markable.head.lemma]:
			return False
		if candidate.head.lemma.isupper() and candidate.head.lemma.lower() in lex.antonyms[markable.head.lemma]:
			return False

	# Check that the heads are not conflicting proper names
	if markable.form == "proper" and candidate.form == "proper":
		if markable.text != candidate.text:
			if markable.text in lex.names and candidate.text in lex.names:
				return False
			elif markable.text.count(" ") == 0 and candidate.text.count(" ") == 0:
				isa = False
				if markable.text in lex.first_names and candidate.text in lex.first_names:
					if markable.text in lex.isa:
						if candidate.text.lower() in lex.isa[markable.text]:
							isa = True
					if candidate.text in lex.isa:
						if markable.text.lower() in lex.isa[candidate.text]:
							isa = True
					if not isa:
						return False
				if markable.text in lex.last_names and candidate.text in lex.last_names:
					if markable.text in lex.isa:
						if candidate.text.lower() in lex.isa[markable.text]:
							isa = True
					if candidate.text in lex.isa:
						if markable.text.lower() in lex.isa[candidate.text]:
							isa = True
					if not isa:
						return False

	# Recursive check through antecedent ancestors in group
	if isinstance(candidate.antecedent, Markable):
		antecedent_compatible = modifiers_compatible(markable, candidate.antecedent, lex)
		if not antecedent_compatible:
			return False

	return True


def agree_compatible(mark1, mark2, lex):
	"""
	Checks if the agree property of two markables is compatible for possible coreference

	:param mark1: the first of two markables to compare agreement
	:param mark2: the second of two markables to compare agreement
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	if mark1.agree == mark2.agree:
		return True
	elif lex.filters["no_person_agree"].match(mark1.agree) and mark2.entity == lex.filters["person_def_entity"]:
		return False
	elif lex.filters["no_person_agree"].match(mark2.agree) and mark1.entity == lex.filters["person_def_entity"]:
		return False
	elif mark1.agree in mark2.alt_agree:
		mark2.agree = mark1.agree
		return True
	elif mark2.agree in mark1.alt_agree:
		mark1.agree = mark2.agree
		return True
	elif (mark1.agree is None or mark1.agree == '') and (mark2.agree is None or mark2.agree == ''):
		return True
	elif (((mark1.agree is None or mark1.agree == '') and lex.filters["agree_with_unknown"].match(mark2.agree) is not None)
	or ((mark2.agree is None or mark2.agree == '') and lex.filters["agree_with_unknown"].match(mark1.agree) is not None)):
		return True
	else:
		return False


def merge_entities(mark1, mark2, previous_markables, lex):
	"""
	Negotiates entity mismatches across coreferent markables and their groups.
	Returns True if merging has occurred.

	:param mark1: the first of two markables to merge entities for
	:param mark2: the second of two markables to merge entities for
	:param previous_markables: all previous markables which may need to inherit from the model/host
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	if not mark1.entity == mark2.entity:
		if mark1.entity in mark2.alt_entities:
			if update_group(mark2, mark1, previous_markables, lex):
				mark2.entity = mark1.entity
				mark2.subclass = mark1.subclass
				return True
			else:
				return False
		else:
			if update_group(mark1, mark2, previous_markables, lex):
				mark1.entity = mark2.entity
				mark1.subclass = mark2.subclass
				return True
			else:
				return False
	else:
		return True


def update_group(host, model, previous_markables, lex):
	"""
	Attempts to update entire coreference group of a host markable with information
	gathered from a model markable discovered to be possibly coreferent with the host.
	If incompatible modifiers are discovered the process fails and returns False.
	Otherwise updating succeeds and the update_group returns true

	:param host: the first markable discovered to be coreferent with the model
	:param model: the model markable, containing new information for the group
	:param previous_markables: all previous markables which may need to inherit from the model/host
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	group = host.group
	for markable in previous_markables:
		if markable.group == group:
			if not modifiers_compatible(markable, model, lex):
				return False
	for markable in previous_markables:
		if markable.group == group:
			markable.entity = model.entity
			markable.subclass = model.subclass
	return True

def isa(markable, candidate, lex):
	"""
	Staging function to check for and store new cached isa information.
	Calls actual :func:`run_isa` function if pair is still viable for new isa match.

	:param markable: one of two markables to compare lexical isa relationship with
	:param candidate: the second markable, which is a candidate antecedent for the other markable
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	retval = False
	if markable.id+"|"+candidate.id not in lex.incompatible_isa_pairs:
		retval = run_isa(markable, candidate, lex)
		if not retval:
			lex.incompatible_isa_pairs.add(markable.id+"|"+candidate.id)
	return retval

def run_isa(markable, candidate, lex):
	"""
	Checks whether two markables are compatible for coreference via the isa-relation

	:param markable: one of two markables to compare lexical isa relationship with
	:param candidate: the second markable, which is a candidate antecedent for the other markable
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""

	if "ablations" in lex.debug:
		if "no_isa" in lex.debug["ablations"]:
			return False

	if not lex.filters["allow_indef_anaphor"]:
		# Don't allow an indefinite to have a definite antecedent via isa relation
		if markable.start > candidate.start:
			if markable.definiteness == "indef" and candidate.definiteness =="def":
				return False
		else:
			if markable.definiteness == "def" and candidate.definiteness =="indef":
				return False

		# Don't allow a proper markable to have an indefinite antecedent via isa relation
		# unless there's corroborating evidence
		#if markable.cardinality == candidate.cardinality and markable.cardinality != 0:
		#	pass  # Explicit cardinality match, forgo indefinite antecedent prohibition
		#elif markable.subclass == candidate.subclass and markable.agree == candidate.agree:
		#	pass  # Explicit subclass and agree match, forgo indefinite antecedent prohibition
		#else: TODO: re-examine indefinite isa antecedents in natural data
		if markable.start > candidate.start:
			if markable.form == "proper" and candidate.definiteness == "indef":
				return False
		else:
			if markable.definiteness == "indef" and candidate.form == "proper":
				return False

	if not lex.filters["allow_indef_isa"]:
		# Don't allow an indefinite to have any antecedent via isa relation if forbidden by configuration
		if markable.start > candidate.start:
			if markable.definiteness == "indef":
				return False
		else:
			if candidate.definiteness =="indef":
				return False

	# Check for incompatible modifiers
	if len(markable.modifiers) > 0:
		if not modifiers_compatible(markable,candidate, lex):
			lex.incompatible_mod_pairs.add(markable.id+"|"+candidate.id)
			return False

	# Check for first name + full name match
	if markable.entity in ["", lex.filters["person_def_entity"]] and candidate.entity in ["", lex.filters["person_def_entity"]]:
		if markable.head.text in lex.first_names:
			candidate_mod_texts = list((mod.text) for mod in candidate.head.modifiers)
			if markable.head.text in candidate_mod_texts:
				return True
		if candidate.head.text in lex.first_names:
			markable_mod_texts = list((mod.text) for mod in markable.head.modifiers)
			if candidate.head.text in markable_mod_texts:
				return True

	# Check for last name + full name match
	if markable.entity in ["", lex.filters["person_def_entity"]] and candidate.entity in ["", lex.filters["person_def_entity"]]:
		if markable.head.text in lex.last_names:
			candidate_mod_texts = list((mod.text) for mod in candidate.head.modifiers)
			if markable.head.text in candidate_mod_texts:
				return True
		if candidate.head.text in lex.last_names:
			markable_mod_texts = list((mod.text) for mod in markable.head.modifiers)
			if candidate.head.text in markable_mod_texts:
				return True

	# Forbid isa head matching for two distinct proper names except first+full name; NB: use coref table for these if desired
	if markable.form == "proper" and candidate.form == "proper":
		if markable.text in lex.names or markable.text in lex.first_names:
			if candidate.text in lex.names or candidate.text in lex.first_names:
				#return False
				pass

	# Subclass based isa match - check agreement too unless disabled
	# Note that this check is unidirectional: the subclass can match an antecedent instance of it,
	# but prior mention of the subclass is not matched to a subsequent instance
	# (the Guardian .. < .. the newspaper is OK, but not: the newspaper .. < .. the Guardian)
	for subclass in candidate.alt_subclasses + [candidate.subclass]:
		if subclass == markable.lemma:
			if agree_compatible(markable, candidate, lex) and not never_agree(markable, candidate, lex):
				# Check if this case is already assigned a different lexical head as isa partner
				if markable.isa_partner_head == "" or markable.isa_partner_head == candidate.lemma:
					candidate.isa_dir = "markable"
					return True
				else:  # Another lemma is already isa-linked to this, e.g. state <- Oregon; so now not also "Nevada"
					return False
		if subclass in lex.isa:
			if lex.isa[subclass][-1] == "*":
				subclass_isa = lex.isa[subclass][:-1]
				check_agree = False
			else:
				subclass_isa = lex.isa[subclass]
				check_agree = lex.filters["isa_subclass_agreement"]
			if markable.lemma.lower() in subclass_isa:
				if markable.isa_partner_head == "" or markable.isa_partner_head == candidate.lemma or candidate.isa_partner_head == markable.lemma:
					if (agree_compatible(markable, candidate, lex) or check_agree is False) and not never_agree(markable, candidate, lex):
						candidate.isa_dir = "markable"
						return True

	# Exact text match in isa table - no agreement matching is carried out
	if markable.text in lex.isa:
		if candidate.text in lex.isa[markable.text]:
			if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
				candidate.isa_dir = "candidate"
				return True
	if candidate.text in lex.isa:
		if markable.text in lex.isa[candidate.text]:
			if markable.isa_partner_head == "" or markable.isa_partner_head == candidate.head.lemma:
				candidate.isa_dir = "markable"
				return True

	# Core text isa match
	# Note this check is unidirectional
	if markable.core_text in lex.isa:
		if candidate.core_text in lex.isa[markable.core_text] or candidate.head.lemma in lex.isa[markable.core_text]:
			if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
				if agree_compatible(markable, candidate, lex) and not never_agree(markable, candidate, lex):
					candidate.isa_dir = "candidate"
					return True
		# Head-core text isa match - no agreement matching is carried out
		elif candidate.head.text in lex.isa[markable.core_text]:
			if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
				candidate.isa_dir = "candidate"
				return True
	elif markable.core_text.isupper():  # Try to title case on all caps entity
		if markable.core_text.title() in lex.isa:
			if candidate.core_text in lex.isa[markable.core_text.title()] or candidate.head.lemma in lex.isa[markable.core_text.title()]:
				if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
					candidate.isa_dir = "candidate"
					return True

	# Handle cases where a prefix like an article is part of the entity name, but a suffix like a possessive isn't
	if remove_suffix_tokens(markable.text,lex) in lex.isa:
		if candidate.head.text in lex.isa[remove_suffix_tokens(markable.text,lex)]:
			if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
				#candidate.isa_partner_head = markable.head.lemma
				candidate.isa_dir = "candidate"
				return True
	elif remove_suffix_tokens(candidate.text, lex) in lex.isa:
		if markable.head.text in lex.isa[remove_suffix_tokens(candidate.text, lex)]:
			if markable.isa_partner_head == "" or markable.isa_partner_head == candidate.head.lemma:
				candidate.isa_dir = "markable"
				return True

	# Head-head isa match - no agreement matching is carried out
	if markable.head.text in lex.isa:
		if candidate.head.text in lex.isa[markable.head.text]:
			if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
				candidate.isa_dir = "candidate"
				return True
	if candidate.head.text in lex.isa:
		if markable.head.text in lex.isa[candidate.head.text]:
			if markable.isa_partner_head == "" or markable.isa_partner_head == candidate.head.lemma:
				candidate.isa_dir = "markable"
				return True

	# Lemma based isa matching - check agreement too
	if markable.head.lemma in lex.isa:
		if candidate.head.lemma in lex.isa[markable.head.lemma] or candidate.head.text in lex.isa[markable.head.lemma]:
			if candidate.isa_partner_head == "" or candidate.isa_partner_head == markable.head.lemma:
				if agree_compatible(markable, candidate, lex):
					candidate.isa_dir = "candidate"
					return True
	if candidate.head.lemma in lex.isa:
		if markable.head.lemma in lex.isa[candidate.head.lemma] or markable.head.text in lex.isa[candidate.head.lemma]:
			if markable.isa_partner_head == "" or markable.isa_partner_head == candidate.head.lemma:
				if agree_compatible(markable, candidate, lex):
					candidate.isa_dir = "markable"
					return True

	return False


def group_agree_compatible(markable,candidate,previous_markables,lex):
	"""
	:param markable: markable whose group the candidate might be joined to
	:param candidate: candidate to check for compatibility with all group members
	:param previous_markables: all previous markables which may need to inherit from the model/host
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""
	if "+" in lex.filters["never_agree_pairs"]:
		never_agreement_pairs = lex.filters["never_agree_pairs"].split(";")
		agreements = []
		for mark in previous_markables:
			if mark.group == markable.group or mark.group == candidate.group:
				agreements.append(mark.agree)

		for pair in never_agreement_pairs:
			class1, class2 = pair.split("+")
			if class1 in agreements and class2 in agreements:
				return False
	return True


def never_agree(candidate, markable, lex):
	if "+" in lex.filters["never_agree_pairs"]:
		never_agreement_list = lex.filters["never_agree_pairs"].split(";")
		never_agreement_pairs = []
		for pair in never_agreement_list:
			never_agreement_pairs.append(pair.split("+"))
		if [markable.agree, candidate.agree] in never_agreement_pairs or [candidate.agree, markable.agree] in never_agreement_pairs:
			return True
	return False


def score_match_heuristic(markable,candidate,features,lex):
	"""
	Basic fall-back function for heuristic match scoring when no classifier is available


	:param makrable:
	:param candidate:
	:param features:
	:return:
	"""

	score = 0 - (markable.sentence.sent_num - candidate.sentence.sent_num)
	# Break ties via proximity
	score -= ((markable.start - candidate.end) * 0.00001 + (markable.start - candidate.start) * 0.000001)
	# Penalize different entity types, but note heuristic only runs if coref_rules allowed this match in the first place
	if markable.entity != candidate.entity:
		score -= 0.5

	if markable.form != "pronoun":
		# Default heuristic for lexical NPs is 'most recent match', so the score is done
		return score

	score += features["d_entidep"]
	if candidate.entity == lex.filters["person_def_entity"]:  # Introduce slight bias to persons
		score += 0.1
	if candidate.entity == lex.filters["subject_func"]:  # Introduce slight bias to subjects
		score += 0.95
	if candidate.agree == markable.agree:  # Slight bias to explicitly identical agreement (not just compatible)
		score += 0.1
	score += features["d_hasa"]

	return score


def best_candidate(markable, candidate_set, lex, rule, take_first=False):
	"""
	:param markable: markable to find best antecedent for
	:param candidate_set: set of markables which are possible antecedents based on some coref_rule
	:param lex: the LexData object with gazetteer information and model settings
	:param propagate: string with feature propagation instructions from coref_rules.tab in lex
	:param rule_num: the rule number of the rule producing the match in coref_rules.tab
	:param clf_name: name of the pickled classifier to use for this rule, or "_default_" to use heuristic matching
	:param take_first: boolean, whether to skip matching and use the most recent candidate (minimum token distance).
				       This saves time if a rule is guaranteed to produce a unique, correct candidate (e.g. reflexives)
	:return: Markable object or None (the selected best antecedent markable, if available)
	"""

	# DEBUG POINT #
	if markable.text == lex.debug["ana"]:
		a=5

	rule_num, clf_name, propagate, score_thresh = rule.rule_num, rule.clf_name, rule.propagation, rule.thresh

	heuristic = True if clf_name == "_default_" or not lex.filters["use_classifiers"] else False

	if len(candidate_set) == 0:
		return None

	candidate_scores = {}
	candidate_features = {}
	score_ranking = {}
	best = None

	for candidate in candidate_set:
		candidate_features[candidate] = markable.extract_features(lex, candidate, candidate_set, dump_position=False)
		candidate_scores[candidate] = score_match_heuristic(markable, candidate, candidate_features[candidate], lex)
		candidate_features[candidate]["rule_num"] = str(rule_num)

	for index, candidate in enumerate(OrderedDict(sorted(candidate_scores.items(), key=lambda x: x[1]))):
		score_ranking[candidate] = 1/float(index+1)

	if lex.dump is not None:  # Only runs if dumping training data

		for candidate in candidate_set:
			dump_features = markable.extract_features(lex, candidate, candidate_set, dump_position=True)
			if lex.dump_headers == []:
				for key in dump_features:
					lex.dump_headers.append(key)
				lex.dump_headers.append("heuristic_score")  # Append the heuristic based score for comparison
				lex.dump_headers.append("rule_num")  # Track rule number for error analysis
			dump_list = []
			for key in dump_features:
				dump_list.append(dump_features[key])
			dump_list = [str(feat) for feat in dump_list]
			dump_list.append(str(score_ranking[candidate]))
			outline = "\t".join(dump_list)
			if outline not in lex.dump_types:
				lex.dump_types.add(outline)
				outline += "\t"+str(rule_num)
				lex.dump.write(outline+"\n")

	max_score = ""

	if take_first:
		markable.matching_rule = str(rule_num)
		return min(candidate_set, key=lambda x: abs(markable.start - x.start))

	clf_input = []
	candidates = []
	for candidate in candidate_scores:
		if lex.dump is None:  # Skip accurate prediction during dump for speed
			clf_input.append((markable, candidate, candidate_set, lex))
		candidates.append(candidate)

	## DEBUG POINT ##
	if markable.text == lex.debug["ana"]:
		a=4

	if lex.dump is None and not heuristic and lex.filters["use_classifiers"]:
		preds = lex.classifiers[clf_name].classify_many(clf_input)
		for i, pred in enumerate(preds):
			candidate_scores[candidates[i]] = pred

	for candidate in candidate_scores:
		if max_score == "":
			max_score = candidate_scores[candidate]
			best = candidate
		elif candidate_scores[candidate] > max_score:
			max_score = candidate_scores[candidate]
			best = candidate

	if max_score < score_thresh:  # The best option is less likely than no coref
		if heuristic:
			pass  # In heuristic mode, an antecedent must always be selected from the set
		else:
			return None  # In classifier mode, if no candidate scores high enough, None is returned

	if not propagate == "nopropagate":
		propagate_entity(markable, best, propagate)
		propagate_agree(markable, best)

	markable.matching_rule = str(rule_num)
	return best


def stems_compatible(verb, noun, lex):
	verb_stem = lex.filters["stemmer_deletes"].sub("",verb.text)
	noun_stem = lex.filters["stemmer_deletes"].sub("",noun.text)
	if verb_stem == noun_stem and len(noun_stem) > 3:
		return True
	if verb.text in lex.nominalizations:
		if noun.text in lex.nominalizations[verb.text]:
			return True
	return False


def acronym_match(mark, candidate, lex):
	"""
	Check whether a Markable's text is an acronym of a candidate Markable's text

	:param mark: The Markable object to test
	:param candidate: The candidate Markable with potentially acronym-matching text
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""
	position = 0
	calibration = 0
	candidate_string = candidate.core_text
	if "ignore_in_acronym" in lex.filters:
		candidate_string = lex.filters["ignore_in_acronym"].sub("", candidate_string)
		candidate_string = candidate_string.replace("  "," ")

	if mark.head.text.isupper() and len(mark.head.text) > 2:
		for word in candidate_string.split(" "):
			if lex.filters["articles"].match(word):
				calibration = -1
			elif len(word) > 0:
				if len(mark.head.text) > position:
					if word[0].isupper() or word == "&":
						if word[0] == mark.head.text[position]:
							position+=1
						else:
							return False
				else:
					return False
		if (position == len(candidate_string.strip().split(" ")) + calibration) and position > 2:
			return True
		else:
			return False
	else:
		return False



"""
Module to generate and run unit tests

Author: Amir Zeldes
"""

from collections import defaultdict
import unittest
import re, os, sys
from .xrenner_xrenner import Xrenner
from .xrenner_coref import find_antecedent

if sys.version_info[0] < 3:
	python_version = 2
else:
	python_version = 3


def generate_test(conll_tokens, markables, parse, model="eng", name="test"):
	tok_count = len(conll_tokens)
	mark_count = 0
	ids = []
	marks_by_id = {}

	# Collect markable groups, assign IDs by extension and count markables
	marks_by_group = defaultdict(list)
	for mark in markables:
		mark_count += 1

		# Assign predictable ID of the form start_end
		id_ = str(mark.start) + "_" + str(mark.end)
		if id_ in ids:
			raise("xrenner generated two markables with same extension: tok" + str(mark.start) + ":tok" + str(mark.end))
		else:
			ids.append(id_)
			mark.id = id_
			marks_by_id[id_] = mark

		marks_by_group[int(mark.group)].append(mark)

	group_count = len(marks_by_group)

	# Serialize group details
	chains = []
	for group in sorted(marks_by_group):
		chain = sorted(marks_by_group[group],key=lambda x: int(x.id[:x.id.find("_")]))
		gid = "g" + chain[0].id
		chain_string = "  "
		for mark in chain:
			chain_string += mark.id + " < "
		chains.append(chain_string[:-3])

	chains.sort(key=lambda x: int(x[2:x.find("_")]))

	snippets = []
	for chain in chains:
		first = chain[2:chain.find("<")-1] if "<" in chain else chain.strip()
		snippet = marks_by_id[first].text[:20] + "..." if len(marks_by_id[first].text) > 20 else marks_by_id[first].text
		snippets.append(snippet)

	zipped = zip(snippets,chains)

	output = ""
	output += "name:" + name + "\n"
	output += "model:" + model + "\n"
	output += "toks:" + str(tok_count) + " # " + " ".join(tok.text for tok in conll_tokens[1:4]) + "..." + "\n"
	output += "marks:" + str(mark_count) + "\n"
	output += "groups:" + str(group_count) + "\n"
	output += "chains:" + "\n"
	for chain in zipped:
		output += "  # " + str(chain[0]) + "\n"
		output += chain[1] + "\n"
	output += "input_data:" + "\n"
	output += "\n".join(parse)
	output += "\n" + "-"*5 + "\n"
	return output


def setUpModule():
	# Read test/tests.dat
	print("\nxrenner unit tests\n" + "=" * 20 + "\nReading test cases from test/tests.dat")
	file = os.path.dirname(os.path.realpath(__file__)) + os.sep + ".." + os.sep + "test" + os.sep + "tests.dat"
	test_data = ""
	with open(file, 'rb') as f:
		test_data = f.read()

	# Populate cases with Case objects
	global cases
	cases = {}
	if python_version < 3:
		case_list = test_data.split("-----")
	else:
		case_list = test_data.decode().split("-----")

	for case in case_list:
		case = case.strip()
		if len(case) > 0:
			case_to_add = Case(case)
			cases[case_to_add.name] = case_to_add

	# Initialize an Xrenner object with the language model and assign to module level xrenner variable for all suites
	print("Initializing xrenner model 'eng'\n")
	global xrenner
	xrenner = Xrenner("eng",override="TEST")
	xrenner.set_doc_name("test_test")


class Test1Model(unittest.TestCase):

	@classmethod
	def setUpClass(cls):
		print("\nTesting model integrity\n" + "-"*30)
		global xrenner
		cls.xrenner = xrenner
		global cases
		cls.cases = cases

	@classmethod
	def tearDownClass(cls):
		global xrenner
		cls.xrenner = xrenner

	def test_model_files(self):
		print("Checking model files:  ")
		# Check that all model components were read and are filled as expected
		self.assertTrue(len(self.xrenner.lex.coref_rules),"check that coref_rules is full")
		self.assertTrue(len(self.xrenner.lex.entities),"check that entities is full")
		self.assertTrue(len(self.xrenner.lex.entity_heads),"check that entity_heads is full")
		self.assertTrue(len(self.xrenner.lex.pronouns),"check that pronouns is full")
		self.assertTrue(len(self.xrenner.lex.filters),"check that filters is full")

		# Optional components, should be included in default model
		self.assertTrue(len(self.xrenner.lex.names),"check that names is full")
		self.assertTrue(len(self.xrenner.lex.stop_list),"check that stop_list is full")
		self.assertTrue(len(self.xrenner.lex.open_close_punct),"check that open_close_punct is full")
		self.assertTrue(len(self.xrenner.lex.open_close_punct_rev),"check that open_close_punct_rev is full")
		self.assertTrue(len(self.xrenner.lex.entity_mods),"check that entity_mods is full")
		self.assertTrue(len(self.xrenner.lex.entity_deps),"check that entity_deps is full")
		self.assertTrue(len(self.xrenner.lex.hasa),"check that hasa is full")
		self.assertTrue(len(self.xrenner.lex.coref),"check that coref is full")
		self.assertTrue(len(self.xrenner.lex.numbers),"check that numbers is full")
		self.assertTrue(len(self.xrenner.lex.affix_tokens),"check that affix_tokens is full")
		self.assertTrue(len(self.xrenner.lex.antonyms),"check that antonyms is full")
		self.assertTrue(len(self.xrenner.lex.isa),"check that isa is full")


class Test2MarkableMethods(unittest.TestCase):

	@classmethod
	def setUpClass(cls):
		print("\n\nRunning markable method tests\n" + "-"*30)
		global xrenner
		cls.xrenner = xrenner
		global cases
		cls.cases = cases
		cls.xrenner.lex.filters["remove_singletons"] = False

	@classmethod
	def tearDownClass(cls):
		global xrenner
		cls.xrenner = xrenner
		cls.xrenner.lex.filters["remove_singletons"] = True

	def test_name(self):
		# Jerry B. Clinton
		print("\nRun markable name test:  ")
		target = self.cases["mark_name_test"]
		self.xrenner.analyze(target.parse.split("\n"), "unittest")
		markables = self.xrenner.markables
		# Check that there are no nested markables in Jerry B. Clinton
		self.assertEqual(len(markables),1)
		# Check that the name is classified as a person
		self.assertEqual(markables[0].entity, "person")

	def test_atomic_mod(self):
		# Israel Machines Corp.
		# Note that Corp. is an organization-marking atomic flagged modifier of the head
		print("\nRun atomic modifier test:  ")
		target = self.cases["mark_atomic_mod_test"]
		self.xrenner.analyze(target.parse.split("\n"), "unittest")
		markables = self.xrenner.markables
		# Check that there are no nested markables in Israel Machines Corp.
		self.assertEqual(len(markables), 1)
		# Check that the name is classified as a person
		self.assertEqual(markables[0].entity, "organization")


class Test3CorefMethods(unittest.TestCase):

	@classmethod
	def setUpClass(cls):

		print("\n\nRunning coref method tests\n" + "-"*30)
		global xrenner
		cls.xrenner = xrenner
		global cases
		cls.cases = cases

	def test_cardinality(self):
		# I saw two birds . The three birds flew .
		print("\nRun cardinality test:  ")
		target = self.cases["cardinality_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(0,result.mark_count,"cardinality test (two birds != the three birds)")

	def test_appos_envelope(self):
		# Meet [[Mark Smith] , [the Governor]]. [He] is the best.
		print("\nRun apposition envelope test:  ")
		target = self.cases["appos_envelope"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"appos envelope test")

	def test_isa(self):
		# I read [the Wall Street Journal]. [That newspaper] is great.
		print("\nRun isa test:  ")
		target = self.cases["isa_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"isa test (Wall Street Journal <- newspaper)")

	def test_hasa(self):
		# The [[CEO] and the taxi driver] ate . [[His] employees] joined them
		print("\nRun hasa test:  ")
		target = self.cases["hasa_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"hasa test (CEO, taxi driver <- his employees)")

	def test_dynamic_hasa(self):
		# Beth was worried about [[Sinead 's] well-being] , and also about Jane . [[Her] well-being] was always a concern .
		print("\nRun dynamic hasa test:  ")
		target = self.cases["dynamic_hasa_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"dynamic hasa test (Sinead 's <- her)")

	def test_entity_dep(self):
		# I have a book , [a dog] and a car. [It] barked.
		print("\nRun entity dep test:  ")
		target = self.cases["entity_dep_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"entity dep test (a book, a dog <- It barked)")

	def test_affix_morphology(self):
		# [A blorker] had a mummelhound in a blargmobile. I saw [the person] .
		print("\nRun affix morphology test:  ")
		target = self.cases["morph_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"affix morph test (A blorker <- the person)")

	def test_verbal_event_stem(self):
		# John [visited] Spain . [The visit] went well .
		print("\nRun verbal event coreference test:  ")
		target = self.cases["verb_test"]
		result = Case(self.xrenner.analyze(target.parse.split("\n"),"unittest"))
		self.assertEqual(target.chains,result.chains,"verbal event stemming (visited <- the visit	)")


class Case:

	def __init__(self, case_string):
		params, parse = case_string.split("input_data:")
		self.parse = parse.strip()
		params = params.replace("\r","")

		self.chains = []
		chain_mode = False

		for line in params.split("\n"):
			line = re.sub(r'#.*','',line).strip()
			if len(line) > 0:
				if chain_mode:
					self.chains.append(line)
				if ":" in line and not chain_mode and not "options" in line:
					key, val = line.split(":")
					if key == "name":
						self.name = val
					elif key == "toks":
						self.tok_count = int(val)
					elif key == "marks":
						self.mark_count = int(val)
					elif key == "groups":
						self.group_count = int(val)
					elif key == "model":
						self.model = val
					elif key == "chains":
						chain_mode = True


def suite():
	# Create test suite
	test_suite = unittest.TestSuite()

	# Add a test case
	test_suite.addTest(unittest.makeSuite(Test1Model))
	test_suite.addTest(unittest.makeSuite(Test2MarkableMethods))
	test_suite.addTest(unittest.makeSuite(Test3CorefMethods))

	return test_suite


def can_be_coreferent(mark1, mark2, lex):
	"""
	Utility function to check whether an xrenner model is capable of finding two markables coreferent

	:param mark1: The :class:`.Markable` object to match to mark2
	:param mark2: The :class:`.Markable` object to match to mark1
	:param lex: the :class:`.LexData` object with gazetteer information and model settings
	:return: bool
	"""

	lex.incompatible_isa_pairs = set([])
	lex.incompatible_mod_pairs = set([])
	prev_markables = [mark1, mark2]
	mark2.sentence.sent_num = 2
	antecedent, propagation = find_antecedent(mark2, prev_markables,lex)
	return antecedent is not None


if __name__ == '__main__':
	xrenner = None
	cases = {}

	test_runner = unittest.TextTestRunner()
	test_runner.run(suite())






from .xrenner_marker import *
from .xrenner_compatible import *
from .xrenner_propagate import *
from .xrenner_rule import CorefRule, ConstraintMatcher

"""
Coreference resolution module. Iterates through markables to find possible matches based on rules.

Author: Amir Zeldes
"""


def find_antecedent(markable, previous_markables, lex, restrict_rule=""):
	"""
	Search for antecedents by cycling through coref rules for previous markables
	
	:param markable: Markable object to find an antecedent for
	:param previous_markables: Markables in all sentences up to and including current sentence
	:param lex: the LexData object with gazetteer information and model settings
	:param restrict_rule: a string specifying a subset of rules that should be checked (e.g. only rules with 'appos')
	:return: candidate, matching_rule - the best antecedent and the rule that matched it
	"""

	# DEBUG point
	if markable.text == lex.debug["ana"]:
		a=5
	candidate = None
	matching_rule = None
	for i, rule in enumerate(lex.coref_rules):
		# If this call of find_antecedent is limited to certain rules, check that the restriction is in the rule
		if restrict_rule == "" or restrict_rule in rule.ana_spec:
			if coref_rule_applies(lex, rule.ana_constraints, markable):
				candidate = search_prev_markables(markable, previous_markables, rule, lex)
				if candidate is not None:
					matching_rule = rule.propagation
					break

	return candidate, matching_rule


def search_prev_markables(markable, previous_markables, rule, lex):
	"""
	Search for antecedent to specified markable using a specified rule
	
	:param markable: The markable object to find an antecedent for
	:param previous_markables: The list of know markables up to and including the current sentence; markables beyond current markable but in its sentence are included for cataphora.
	:param ante_constraints: A list of ContraintMatcher objects describing the antecedent
	:param ante_spec: The antecedent specification part of the coref rule being checked, as a string
	:param lex: the LexData object with gazetteer information and model settings
	:param max_dist: Maximum distance in sentences for the antecedent search (0 for search within sentence)
	:param propagate: Whether to progpagate features upon match and in which direction
	:return: the selected candidate Markable object
	"""

	ante_constraints, ante_spec, rule_num, max_dist, propagate, clf_name = rule.ante_constraints, rule.ante_spec, rule.rule_num, rule.max_distance, rule.propagation, rule.clf_name

	candidate_set = set([])
	if ante_spec.find("lookahead") > -1:
		referents_to_loop = previous_markables
	else:
		referents_to_loop = reversed(previous_markables)
	for candidate in referents_to_loop:  # loop through previous markables backwards

		#DEBUG breakpoint:
		if markable.text == lex.debug["ana"]:
			a = 5
			if candidate.text == lex.debug["ante"]:
				b=6
		if markable.sentence.sent_num - candidate.sentence.sent_num <= max_dist:
			if ((int(markable.head.id) > int(candidate.head.id) and
			ante_spec.find("lookahead") == -1) or (int(markable.head.id) < int(candidate.head.id) and ante_spec.find("lookahead") > -1)):
				if candidate.group not in markable.non_antecdent_groups:
					if coref_rule_applies(lex, ante_constraints, candidate, markable):
						if not markables_overlap(markable, candidate, lex):
							if markable.form == "pronoun":
								if agree_compatible(markable, candidate, lex) or (ante_spec.find("anyagree") > -1 and group_agree_compatible(markable,candidate,previous_markables,lex)):
									if entities_compatible(markable, candidate, lex) and cardinality_compatible(markable, candidate, lex):
										candidate_set.add(candidate)
							elif markable.text == candidate.text or (len(markable.text) > 4 and (candidate.text.lower() == markable.text.lower())):
								#propagate_entity(markable, candidate, propagate)
								candidate_set.add(candidate)
								#return candidate
							elif markable.text + "|" + candidate.text in lex.coref and entities_compatible(
									markable, candidate, lex) and agree_compatible(markable, candidate, lex):
								candidate_set.add(candidate)
								#return candidate
							elif markable.core_text + "|" + candidate.core_text in lex.coref and entities_compatible(
									markable, candidate, lex) and agree_compatible(markable, candidate, lex):
								candidate_set.add(candidate)
								#return candidate
							elif markable.entity == candidate.entity and agree_compatible(markable, candidate, lex) and (markable.head.text == candidate.head.text or
							(len(markable.head.text) > 3 and (candidate.head.text.lower() == markable.head.text.lower())) or
							(markable.core_text.count(" ") > 2 and (markable.core_text.lower() == candidate.core_text.lower())) or
							(markable.head.lemma == candidate.head.lemma and lex.filters["lemma_match_pos"].match(markable.head.pos) is not None
							and lex.filters["lemma_match_pos"].match(candidate.head.pos) is not None)):
								if modifiers_compatible(markable, candidate, lex) and modifiers_compatible(candidate, markable, lex):
									candidate_set.add(candidate)
							elif (markable.entity == candidate.entity or len(set(markable.alt_entities) & set(candidate.alt_entities))>0) and isa(markable, candidate, lex):
								candidate.isa = True  # This is an 'isa' candidate
								candidate_set.add(candidate)
							elif agree_compatible(markable,candidate,lex) and ((markable.head.text == candidate.head.text) or (markable.head.lemma == candidate.head.lemma and
							lex.filters["lemma_match_pos"].match(markable.head.pos) is not None and lex.filters["lemma_match_pos"].match(candidate.head.pos) is not None)):
								if merge_entities(markable, candidate, previous_markables, lex):
									candidate_set.add(candidate)
							elif entities_compatible(markable, candidate, lex) and isa(markable, candidate, lex):
								if merge_entities(markable, candidate, previous_markables, lex):
									candidate.isa = True  # This is an 'isa' candidate
									candidate_set.add(candidate)
						elif lex.filters["match_acronyms"] and markable.head.text.isupper() or candidate.head.text.isupper():
								if acronym_match(markable, candidate, lex) or acronym_match(candidate, markable, lex):
									if modifiers_compatible(markable, candidate, lex) and modifiers_compatible(candidate, markable, lex):
										if merge_entities(markable, candidate, previous_markables, lex):
											candidate_set.add(candidate)
						if ante_spec.find("anytext") > -1:
								if (ante_spec.find("anyagree") > -1 and group_agree_compatible(markable,candidate,previous_markables,lex)) or agree_compatible(markable, candidate, lex):
									if (ante_spec.find("anycardinality") > -1 or cardinality_compatible(markable,candidate,lex)):
										if (ante_spec.find("anyentity") > -1 or entities_compatible(markable,candidate,lex)):
											candidate_set.add(candidate)
		elif ante_spec.find("lookahead") == -1:
			# Reached back too far according to max_dist, stop looking
			break

	if len(candidate_set) > 0:
		candidates_to_remove = set([])
		for candidate_item in candidate_set:
			# Remove items that are prohibited by entity agree mapping
			for agree, ent in iteritems(lex.filters["agree_entity_mapping"]):
				if markable.agree == agree and candidate_item.entity != ent:
					candidates_to_remove.add(candidate_item)
			if candidate_item.entity == lex.filters["person_def_entity"] and (candidate_item.form != "pronoun" or markable.entity_certainty == "certain") and lex.filters["no_person_agree"].match(markable.agree) is not None:
				candidates_to_remove.add(candidate_item)
			elif markable.entity == lex.filters["person_def_entity"] and (markable.form != "pronoun" or markable.entity_certainty == "certain") and lex.filters["no_person_agree"].match(candidate_item.agree) is not None:
				candidates_to_remove.add(candidate_item)

		for removal in candidates_to_remove:
			candidate_set.remove(removal)

		if len(candidate_set) > 0:
			take_first = True if ante_spec.find("takefirst") > -1 else False
			best = best_candidate(markable, candidate_set, lex, rule, take_first=take_first)
			if best is not None:
				if markable.text + "|" + best.text in lex.coref:
					markable.coref_type = lex.coref[markable.text + "|" + best.text]
					propagate_entity(markable, best, propagate)
					propagate_entity(markable, best)
				elif markable.core_text + "|" + best.core_text in lex.coref:
					markable.coref_type = lex.coref[markable.core_text + "|" + best.core_text]
					propagate_entity(markable, candidate_item)
				elif propagate.startswith("propagate"):
					propagate_entity(markable, best, propagate)
			if hasattr(best,"isa"):
				if hasattr(best,"isa_dir"):
					if best.isa_dir == "markable":
						markable.isa_partner_head = best.lemma
					else:
						best.isa_partner_head = markable.lemma
					delattr(best,"isa_dir")
				delattr(best,"isa")
			return best
		else:
			return None
	else:
		return None


def coref_rule_applies(lex, constraints, mark, anaphor=None):
	"""
	Check whether a markable definition from a coref rule applies to this markable
	
	:param lex: the LexData object with gazetteer information and model settings
	:param constraints: the constraints defining the relevant Markable
	:param mark: the Markable object to check constraints against
	:param anaphor: if this is an antecedent check, the anaphor is passed for $1-style constraint checks
	:return: bool: True if 'mark' fits all constraints, False if any of them fail
	"""
	for constraint in constraints:
		if not constraint.match(mark,lex,anaphor):
			return False
	return True


def antecedent_prohibited(markable, conll_tokens, lex):
	"""
	Check whether a Markable object is prohibited from having an antecedent
	
	:param markable: The Markable object to check
	:param conll_tokens: The list of ParsedToken objects up to and including the current sentence
	:param lex: the LexData object with gazetteer information and model settings
	:return: bool
	"""
	mismatch = True
	if "/" in lex.filters["no_antecedent"]:
		constraints = lex.filters["no_antecedent"].split(";")
		for constraint in constraints:
			if not mismatch:
				return True
			descriptions = constraint.split("&")
			mismatch = False
			for token_description in descriptions:
				if token_description.startswith("^"):
					test_token = conll_tokens[markable.start]
				elif token_description.startswith("$"):
					test_token = conll_tokens[markable.end]
				elif token_description.startswith("@"):
					test_token = markable.head
				else:
					# Invalid token description
					return False
				token_description = token_description[1:]
				pos, word = token_description.split("/")
				if pos.startswith("!"):
					pos = pos[1:]
					negative_pos = True
				else:
					negative_pos = False
				if word.startswith("!"):
					word = word[1:]
					negative_word = True
				else:
					negative_word = False
				pos_matcher = re.compile(pos)
				word_matcher = re.compile(word)
				if (pos_matcher.match(test_token.pos) is None and not negative_pos) or (pos_matcher.match(test_token.pos) is not None and negative_pos) or \
				(word_matcher.match(test_token.text) is None and not negative_word) or (word_matcher.match(test_token.text) is not None and negative_word):
					mismatch = True
					break
	if mismatch:
		return False
	else:
		return True






#!/usr/bin/python
# -*- coding: utf-8 -*-


def toks_to_sents(text, split_at_punct=True):

	lines = text.split("\n")
	output = []
	if not lines[0] == "<s>":
		output.append("<s>")

	for i, line in enumerate(lines):
		if len(output) == 0:
			output.append(line)
		else:
			if not (line == "<s>" and output[-1] == "<s>") and not (line == "</s>" and output[-1] == "<s>"):
				output.append(line)
		if line in [".","!","?"] and split_at_punct:
			output.append("</s>")
			output.append("<s>")

	if output[-1] != "</s>":
		output.append("</s>")

	return "\n".join(output)



"""
Postprocessing module. Alters results of coreference analysis based on model settings,
such as deleting certain markables or re-wiring coreference relations according to a particular
annotation scheme

Author: Amir Zeldes and Shuo Zhang
"""

from collections import defaultdict
from .xrenner_classes import *
from .xrenner_marker import markables_overlap, markable_extend_punctuation


def postprocess_coref(markables, lex, markstart, markend, markbyhead, conll_tokens):


	# Collect markable groups
	marks_by_group = defaultdict(list)
	for markable in markables:
		if markable.antecedent != "none":
			if markable.group != markable.antecedent.group:
				markable.group = markable.antecedent.group
		marks_by_group[markable.group].append(markable)

	# Order markables in each group to ensure backwards chain
	# except in the case of cataphors, which point forward
	for group in marks_by_group:
		last_mark = None
		for mark in marks_by_group[group]:
			if mark.coref_type != "cata":
				if last_mark is not None:
					mark.antecedent = last_mark
				last_mark = mark

	# Check for markables to remove in postprocessing
	if len(lex.filters["remove_head_func"].pattern) > 0:
		for mark in markables:
			if lex.filters["remove_head_func"].match(mark.head.func) is not None and (mark.form != "proper" or
						mark.entity == "abstract" or
						(mark.text in ["U.S.","US"] and mark.func == "nn") or (mark.text in lex.first_names and mark.entity != lex.filters["time_def_entity"])): # TODO: de-hardwire Proper restriction matching OntoNotes guidelines; US is interpreted as "American" (amod); forbid abstract nn modifier even if proper
				splice_out(mark, marks_by_group[mark.group])
	if len(lex.filters["remove_child_func"].pattern) > 0:
		for mark in markables:
			for child_func in mark.head.child_funcs:
				if lex.filters["remove_child_func"].match(child_func) is not None and mark.head.func != "cata":
					splice_out(mark, marks_by_group[mark.group])
	# Handle nested compound persons
	if len(lex.rm_nested_entities) > 0:
		for mark in markables:
			if remove_nested(mark,markbyhead,lex):
				splice_out(mark, marks_by_group[mark.group])

	# Remove i in i rule (no overlapping markable coreference in group)
	# TODO: make this more efficient (iterates all pairwise comparisons)
	for group in marks_by_group:
		for mark1 in marks_by_group[group]:
			for mark2 in marks_by_group[group]:
				if not mark1 == mark2:
					if markables_overlap(mark1, mark2, None):
						if (mark1.end - mark1.start) > (mark2.end - mark2.start):
							splice_out(mark2, marks_by_group[group])
						else:
							splice_out(mark1, marks_by_group[group])

	# Remove cataphora if desired
	if lex.filters["remove_cataphora"]:
		for mark in markables:
			if mark.coref_type == "cata":
				mark.id = "0"
				if mark.antecedent != "none":
					mark.antecedent.id = "0"

	# Remove coordination envelopes if desired
	if lex.filters["remove_coordinate_envelopes"]:
		for group in marks_by_group:
			coordination_text = ""
			wipe_coord = False
			for mark in marks_by_group[group]:
				if mark.coordinate:
					coordination_text = mark.core_text
					wipe_coord = True
			if coordination_text != "":
				for mark in marks_by_group[group]:
					if mark.core_text != coordination_text:  # This group has multiple text realizations
						wipe_coord = False
			if wipe_coord:
				for mark in marks_by_group[group]:
					mark.id = "0"

	# Inactivate singletons if desired by setting their id to 0
	if lex.filters["remove_singletons"]:
		for group in marks_by_group:
			wipe_group = True
			if len(marks_by_group[group]) < 2:
				for singleton in marks_by_group[group]:
					singleton.id = "0"
			else:
				for singleton_candidate in marks_by_group[group]:
					if singleton_candidate.antecedent != 'none':
						wipe_group = False
						break
				if wipe_group:
					for singleton in marks_by_group[group]:
						singleton.id = "0"


	# Add apposition envelopes if desired
	if lex.filters["add_appos_envelopes"]:
		for group in marks_by_group:
			for i in reversed(range(1,len(marks_by_group[group]))):
				# Print marks_by_group[group]
				mark = marks_by_group[group][i]
				prev = mark.antecedent
				if prev != "none":
					if prev.coref_type == "appos" and prev.antecedent != "none":
						# Two markables in the envelop: prev and prevprev
						prevprev = prev.antecedent
						envlop = create_envelope(prevprev, prev, conll_tokens)

						# Extend markable to trailing closing punctuation if it contains opening punctuation
						if envlop.end < len(conll_tokens) - 1:
							next_id = envlop.end + 1
							lex.open_close_punct[","] = "," # Add trailing comma option to envelope to match OntoNotes behavior
							if markable_extend_punctuation(envlop.text, conll_tokens[next_id], lex.open_close_punct, "trailing"):
								envlop.text += conll_tokens[next_id].text + " "
								envlop.end += 1
							# Idiosyncratic "years old" behavior
							elif conll_tokens[envlop.end].text == "years" and conll_tokens[next_id].text == "old":
								envlop.text += conll_tokens[next_id].text + " "
								envlop.end += 1

						markables.append(envlop)
						markstart[envlop.start].append(envlop)
						markend[envlop.end].append(envlop)

						# Markables_by_head
						head_id=str(prevprev.head.id) + "_" + str(prev.head.id)
						markbyhead[head_id] = envlop

						# Set some fields for the envlop markable
						envlop.non_antecdent_groups = prev.antecedent
						# New group number for the two markables inside the envelope
						ab_group = 1000 + int(prevprev.group) + int(prev.group)
						prevprev.group = ab_group
						prev.group = ab_group
						mark.antecedent = envlop
						prevprev.antecedent = "none"

	kill_zero_marks(markables, markstart, markend)


def kill_zero_marks(markables, markstart_dict, markend_dict):
	"""
	Removes markables whose id has been set to 0 in postprocessing
	
	:param markables: All Markable objects
	:param markstart_dict: Dictionary of token span start ids to lists of markables starting at that id
	:param markend_dict: Dictionary of token span end ids to lists of markables ending at that id
	:return: void
	"""
	marks_to_kill = []
	for mark in markables:
		if mark.id == "0":  # Markable has been marked for deletion
			markstart_dict[mark.start].remove(mark)
			if len(markstart_dict[mark.start]) < 1:
				del markstart_dict[mark.start]
			markend_dict[mark.end].remove(mark)
			if len(markend_dict[mark.start]) < 1:
				del markend_dict[mark.start]
			marks_to_kill.append(mark)

	for mark in marks_to_kill:
		markables.remove(mark)


def splice_out(mark, group):
	min_id = 0
	mark_id = int(mark.id.replace("referent_", ""))
	for member in group:
		if member.antecedent == mark:
			member.antecedent = mark.antecedent
		member_id = int(member.id.replace("referent_", ""))
		if (min_id == 0 or min_id > member_id) and member.id != mark.id:
			min_id = member_id
	mark.antecedent = "none"
	if str(mark_id) != mark.group:
		mark.group = str(mark_id)
	else:
		for member in group:
			if member != mark:
				member.group = str(min_id)
	mark.id = "0"


def create_envelope(first,second, conll_tokens):
	mark_id="env"
	form = "proper" if (first.form == "proper" or second.form == "proper") else "common"
	head=first.head
	definiteness=first.definiteness
	start=first.start
	end=second.end
	intermediate = ""
	if first.end+1 < second.start: # Some intervening tokens should be included in the envelope text
		for tok in conll_tokens[first.end+1:second.start]:
			intermediate += tok.text + " "
	text=first.text.strip() + " " + intermediate + second.text.strip()
	entity=second.entity
	entity_certainty=second.entity_certainty
	subclass=first.subclass
	infstat=first.infstat
	agree=first.agree
	sentence=first.sentence
	antecedent=first.antecedent
	coref_type=first.coref_type
	group=first.group
	alt_entities=first.alt_entities
	alt_subclasses=first.alt_subclasses
	alt_agree=first.alt_agree
	cardinality=0
	if first.cardinality!=0:
		if first.cardinality == second.cardinality:
			cardinality = first.cardinality

	envelope = Markable(mark_id, head, form, definiteness, start, end, text, text, entity, entity_certainty, subclass, infstat, agree, sentence, antecedent, coref_type, group, alt_entities, alt_subclasses, alt_agree, cardinality)

	return envelope


def remove_nested(mark,markbyhead,lex):
	for nested_entity, func, container_entity in lex.rm_nested_entities:
		if mark.func == func and mark.entity == nested_entity:
			if mark.head.head in markbyhead:
				container = markbyhead[mark.head.head]
				if container.entity == container_entity:
					return True
	return False



import os, sys, re

lines = open("wiki_corpus.conllu",encoding="utf8").read().split("\n")

output = []
meta = []
sid = 1
for line in lines:
    if line.startswith("#"):
        if "newdoc_id" in line:
            line = line.replace("newdoc_id","newdoc id")
            output.append(line)
            continue
        elif line.startswith("#"):
            if "= NULL" in line:
                line = line.replace("= NULL","")
            line = line.strip().replace("&quot;",'"').replace("&amp;","&").replace("&gt;",">").replace("&lt;","<")
            meta.append(line)
    elif line.startswith("1\t") or line.startswith("1-"):
        if len(meta)>0:
            meta = ["# sent_id = wiki_corpus-" + str(sid)] + meta
            output += sorted(meta)
            sid += 1
            meta = []
        output.append(line)
    else:
        output.append(line)

with open("wiki_corpus_clean.conllu",'w',encoding="utf8",newline="\n") as f:
    f.write("\n".join(output).strip() + "\n\n")


tags = {"NOUN","VERB","ADJ","ADV","PROPN"}

lex = open("C:\\Uni\\Hebrew\\HebPipe.git\\hebpipe\\data\\heb.lex",encoding="utf8").read().strip().split("\n")
lex = [l for l in lex if "\t" in l]
lex = [l for l in lex if l.split("\t")[1] in tags]

with open("heb.lemma",'w',encoding="utf8",newline="\n") as f:
    f.write("\n".join(sorted(lex)))